#include "PedSeq.h"

#include <cassert>
#include <string>
#include <set>
#include <map>
#include <vector>
#include <algorithm>
#include <iostream>
#include "Argument.h"
#include "IO.h"
#include "tabix.h"
#include "PeopleSet.h"
#include "RangeList.h"
#include "Utils.h"
#include "VCFUtil.h"
#include "Utils.h"

PedSeq::PedSeq(){
  GL_idx = -1;
  PL_idx = -1;
  DP_idx = -1;
  prior = 0.;
  prior_indel = 0.0;
  isIndel = false;
  nFounders = 0;
  output_all = false;
  phase_marginal = false;
  phase_best = false;
  geno_label.resize(3);
  geno_label[0] = "0/0";
  geno_label[1] = "0/1";
  geno_label[2] = "1/1";
  out_chr = "";
  out_start = 0;
  out_end = 0;
  onfly = false;
  cumulative_marginals = 1.0;
}

PedSeq::~PedSeq() {
  for(int i=0; i<pm->nFam; i++)
    delete [] single_marker[i];
  delete [] single_marker;

  delete [] famlk;
  delete [] im;
}

double PedSeq::CalculatePosteriorFromLLK(std::vector<double> &llk)
{
  if(llk.size()==0) error("Log lk is emptye!\n");

  int max_idx = 0;
  double max_llk = llk[0];

  for(int i=1; i<llk.size(); i++) if(max_llk<llk[i]) { max_llk = llk[i]; max_idx = i; }

  std::vector<double> ratio;
  ratio.resize(llk.size());
  double ratio_sum = 0.0;

  for(int i=0; i<ratio.size(); i++){
    ratio[i] = pow10(llk[i]-llk[max_idx]);
  }

  for(int i=0; i<ratio.size(); i++) ratio_sum += ratio[i];

  double post = 1/ratio_sum;

  return(post);
}

double PedSeq::CalculatePosteriorFromLLK(std::vector<double> &llk, int post_idx)
{
  if(llk.size()==0) error("Log lk is emptye!\n");
  if(post_idx<0 || post_idx>=llk.size()) error("Posterior index out of range!\n");

  int max_idx = 0;
  double max_llk = llk[0];

  for(int i=1; i<llk.size(); i++) if(max_llk<llk[i]) { max_llk = llk[i]; max_idx = i; }

  std::vector<double> ratio;
  ratio.resize(llk.size());
  double ratio_sum = 0.0;

  for(int i=0; i<ratio.size(); i++){
    ratio[i] = pow10(llk[i]-llk[max_idx]);
  }

  for(int i=0; i<ratio.size(); i++) ratio_sum += ratio[i];

  double post = ratio[post_idx]/ratio_sum;

  return(post);
}

bool PedSeq::isTransition(std::string ref, std::string alt)
{
  if(ref.size()>1 || alt.size()>1) return(false);
  if(toupper(ref[0])=='A' && toupper(alt[0])=='G' || toupper(ref[0])=='G' && toupper(alt[0])=='A' ) return(true);
  if(toupper(ref[0])=='C' && toupper(alt[0])=='T' || toupper(ref[0])=='T' && toupper(alt[0])=='C' ) return(true);
  return(false);
}

void PedSeq::SetPedMarkerData(PedMarkerData *p){
  pm = p;
  im = new Inheritance_Matrix[pm->nFam];
  famlk = new FamilyLikelihoodLG[pm->nFam];
  single_marker = new SeqMarkerInfo*[pm->nFam];
  for(int i=0; i<pm->nFam; i++) {
    single_marker[i] = new SeqMarkerInfo[pm->ped->families[i]->count];
    famlk[i].SetCumulativeMarginals(cumulative_marginals);
  }
}

void PedSeq::SetPolyPrior()
{
  prior = 0;
  prior_indel = 0;
  nFounders = 0;

  for(int i=0; i<pm->nFam; i++)
    {
      Family *fam;
      fam= pm->ped->families[i];
      nFounders += im[i].nFounders;
     }

  if(nFounders==0)
    error("Family size is zero\n");

  for(int i=1; i<=2*nFounders; i++) //number of independent chromosomes
    prior += 1.0 / i;

  prior *= theta;
  prior_indel *= theta_indel;
}

void PedSeq::SetSNVPrior(std::string &ref, std::string &alt)
{
 if(isTransition(ref, alt))
  snv_prior = prior * tstv_ratio/(tstv_ratio + 1);
 else snv_prior = prior * 0.5/(tstv_ratio+1);
}

void PedSeq::SetInheritanceMatrix()
{
  for(int i=0; i<pm->nFam; i++)
    {
      Family *fam;
      fam= pm->ped->families[i];

      im[i].SetPedigree(pm->ped);
      im[i].SetFamily(fam);
      im[i].Create_Inheritance_Matrix();
      if(onfly) im[i].SetFounderNode();
      else im[i].MapFounderNode();
      im[i].SetZeroBits();

      famlk[i].SetOnflyFlag(onfly);
      famlk[i].SetPedMarkerData(pm, i);
      famlk[i].SetInheritanceMatrix(&im[i]);
      famlk[i].AllocateMemory();
    }
}

void PedSeq::CalculateIVMarginals()
{
  #pragma omp parallele for
  for(int i=0; i<pm->nFam; i++)
    {
      printf("Calculating inheritance vector probabilities ...\n");
      famlk[i].CalculateIVProb();

      printf("Calculating multiple point likelihoood: forward ...\n");
      if(!disable_fft) famlk[i].CalculateMultiPointLikelihood_forward_FFT();
      else famlk[i].CalculateMultiPointLikelihood_forward();

      printf("Calculating multiple point likelihoood: backword ...\n");
      if(!disable_fft) famlk[i].CalculateMultiPointLikelihood_backward_FFT();
      else famlk[i].CalculateMultiPointLikelihood_backward();

      printf("back done\n");

      famlk[i].CalculateMarginals();
      famlk[i].FindBestMarginal();

      if(phase_best)
        famlk[i].FindBestPath();

      famlk[i].Clear_Partial();
    }
}

void PedSeq::CalculateGenoPosteriorAndOutput2VCF()
{
  int counter = 0;
  int lpos = -1;
  int rpos = -1;
  int lidx = -1;
  int ridx = 0;
  int max_idx = 0;
  double max_lk_value = 0.0;
  double lk_l = -1.0;
  double lk_r = -1.0;

  std::vector<int> max_lk_idx; //record the position index of loc with larger lk on two boundariefor varinats in between
  max_lk_idx.resize(pm->nFam);
  std::vector<int> max_lk;  // record the lk which is larger of the two in two boundaries for variants in between
  max_lk.resize(pm->nFam);

  FILE *fh = fopen(vcfOutFile.c_str(), "w");
  if(fh==NULL) error("Open output vcf file failed!\n");

  std::string header = "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT";
  std::pair<int, int> fam_pid_index_pair;

  for(int i=0; i<pm->nFam; i++)
  for(int j=0; j<pm->ped->families[i]->count; j++)
  {
    fam_pid_index_pair.first = i;
    fam_pid_index_pair.second = j;

    if(output_all==false && pm->fam_pid_index_pair_with_data.count(fam_pid_index_pair)==0) continue;

    header += "\t";
    header += pm->traverse2pid[fam_pid_index_pair];
  }

  std::string meta_data = "##fileformat=VCFv4.1\n";
  meta_data += "##Polymutt2=";
  meta_data += cmd;
  // Here needs to insert the command
  meta_data += "\n";
  meta_data = meta_data +"##Note=VCF file modified by polymutt2. Updated fileds include: QUAL, GT and GQ, and AC. NOTE: modification was applied only to biallelic variants\n" +
    "##FILTER=<ID=LOWDP,Description=\"Low Depth filter when the average depth per sample is lessn than 1\">\n" +
    "##INFO=<ID=DP,Number=1,Type=Integer,Description=\"Total Read Depth\">\n" +
    "##INFO=<ID=AF,Number=A,Type=Float,Description=\"Alternative Allele Frequency\">\n" +
    "##INFO=<ID=AC,Number=1,Type=Integer,Description=\"Alternative Allele Count\">\n" +
    "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n" +
    "##FORMAT=<ID=GQ,Number=1,Type=Integer,Description=\"Genotype Quality\">\n" +
    "##FORMAT=<ID=DP,Number=1,Type=Integer,Description=\"Read Depth\">\n";

    if(phase_best) meta_data += "##FORMAT=<ID=BI,Number=2,Type=Integer,Description=\"Best path inheritance vector. Founder alleles are arbiturally labeled (1 to 2*nFounders) and L1|L2 for non-founders indicated L1 and L2 from founders are transmitted\">\n";
    else meta_data += "##FORMAT=<ID=MI,Number=2,Type=Integer,Description=\"Marginal inheritance vector. Founder alleles are arbiturally labeled (1 to 2*nFounders) and L1|L2 for non-founders indicated L1 and L2 from founders are transmitted\">\n";

    meta_data = meta_data + "##FORMAT=<ID=PL,Number=3,Type=Integer,Description=\"Phred-scaled Genotype Likelihoods\">\n" +
    "##FORMAT=<ID=GL,Number=3,Type=Float,Description=\"Log10 Genotype Likelihoods\">\n" +
    header + "\n";
  fprintf(fh, "%s", meta_data.c_str() );

  printf("Inferring genotypes and writing to %s...\n", vcfInFile.c_str());
  VCFInputFile vin(vcfInFile.c_str());
  vin.includePeople(pm->pids);

for(int p=0; p<pm->pids.size(); p++) printf("Including pid in PedSeq: %s\n", pm->pids[p].c_str());

  PL_idx = -1;
  GL_idx = -1;
  DP_idx = -1;

  int cnt=0;
  while(vin.readRecord())
    {
      VCFRecord &r = vin.getVCFRecord();
      VCFPeople &people = r.getPeople();

      if(people.size()==0) error("No individuals are present in the VCF file!\n");

      single_var.chr = r.getChrom();
      if(single_var.chr.substr(0,3).compare("chr")==0) { single_var.chr = single_var.chr.substr(3, single_var.chr.size()); chr_prefix = "chr"; }
      single_var.pos = r.getPos();
      single_var.ref = r.getRef();
      single_var.alt = r.getAlt();
      single_var.ac = 0;
      single_var.dp = 0;

      if(!out_chr.empty()) {
        if(single_var.chr.compare(out_chr) !=0) continue;
        if(single_var.pos<out_start || single_var.pos>out_end) continue;
      }

       //Skip variants called as the ref allele
       if(single_var.ref == single_var.alt) continue;

      //Check ref and alt alleles, and skip if not bi-allelic
       SetBiallelic(true);
       SetIndel(false);

       if(single_var.alt.find(',')!=string::npos) 
	SetBiallelic(false);
       else{
	SetBiallelic(true);
        if(single_var.ref.size()>1 || single_var.alt.size()>1) SetIndel(true);
       }

       if(isBiallelic==false) continue;

       SetSNVPrior(single_var.ref, single_var.alt); // Set SNV prior for the REF and ALT pair based on user input ts/tv ratio

      // Read AF value from INFO
      bool missing=false;
      single_var.freq = 1 - r.getInfoTag("AF", &missing).toDouble(); // here it assumes AF is the ALT frequency
      if(single_var.freq>=1.0) single_var.freq = 0.999;
      if(single_var.freq<=0.0) single_var.freq = 0.001;

      if(missing) error("AF is not in the INFO in the VCF file at %s:%d! You can run POLYMUTT to calculate AF values\n", single_var.chr.c_str(), single_var.pos);

      // Read DP value from INFO
      //single_var.dp = r.getInfoTag("DP", &missing).toInt(); // here it assumes AF is the ALT frequency
      //if(missing) single_var.dp = 0;

      if(pm->chr.size()>0)
        {
	  if(pm->chr.compare(single_var.chr)!=0) continue;
	  if(pm->end>0)
	    if(r.getPos()<pm->start) continue;
	    else if(r.getPos()>pm->end) break;
        }

      // Here GL and PL are checked only once. The index needs to be fixed for all records
      if(GL_idx<0 && PL_idx<0)
	{
	  GL_idx = r.getFormatIndex("GL");
	  PL_idx = r.getFormatIndex("PL");
	  DP_idx = r.getFormatIndex("DP");

	  if(GL_idx<0 && PL_idx<0) {
	    fprintf(stderr, "NO GL field was found. Please check the vcf file at chr:%s and position:%d", single_var.chr.c_str(), r.getPos());
	    exit(1);
	  }
	}

      // Read in VCF records for all individuals and fill in penetrance matrix
      VCFIndividual* indv;
      std::map<std::string, std::pair<int, int> >::iterator it;
      StringArray gl;

      for(int i=0; i<people.size(); i++)
        {
	  indv = people[i];
	  it = pm->pid2traverse.find( (*indv).getName() );
	  if(it==pm->pid2traverse.end() ) continue;

	  bool missing = false;
	  std::string pl_str = GL_idx>0 ? (*indv).get(GL_idx, &missing).toStr() : (*indv).get(PL_idx, &missing).toStr(); //Bug: seg fault is the PL value is '.' but why it is ok for missing depth?

          if(missing || pl_str.compare(".")==0) { missing = true; }
	  else
          {
           if(pl_str.find(",")==string::npos) error("PL format seems wrong!\n");

	   gl.ReplaceTokens(String(pl_str.c_str()), ",");

           if(gl.Length()<3) missing = true;
	   else if(gl[0]=="." || gl[1]=="." || gl[2]==".") missing = true;
	   else missing = false;
           }

          single_marker[it->second.first][it->second.second].missing_pl = missing;

	  if(missing) {
	    single_marker[it->second.first][it->second.second].geno_lk[0] = 1.0;
	    single_marker[it->second.first][it->second.second].geno_lk[1] = 1.0;
	    single_marker[it->second.first][it->second.second].geno_lk[2] = 1.0;
	    single_marker[it->second.first][it->second.second].geno_pl[0] = 0;
	    single_marker[it->second.first][it->second.second].geno_pl[1] = 0;
	    single_marker[it->second.first][it->second.second].geno_pl[2] = 0;
	  }
	  else{
	  double GL0 = gl[0].AsDouble();
	  double GL1 = gl[1].AsDouble();
	  double GL2 = gl[2].AsDouble();

	  int PL0, PL1, PL2;
	  PL0 = int(PL_idx>0 ? GL0 : -10*GL0); if(PL0>255) PL0 = GLLIM;
	  PL1 = int(PL_idx>0 ? GL1 : -10*GL1); if(PL1>255) PL1 = GLLIM;
	  PL2 = int(PL_idx>0 ? GL2 : -10*GL2); if(PL2>255) PL2 = GLLIM;


	    single_marker[it->second.first][it->second.second].geno_lk[0] = pmd.PL2LK(PL0);
	    single_marker[it->second.first][it->second.second].geno_lk[1] = pmd.PL2LK(PL1);
	    single_marker[it->second.first][it->second.second].geno_lk[2] = pmd.PL2LK(PL2);
	    single_marker[it->second.first][it->second.second].geno_pl[0] = PL0;
	    single_marker[it->second.first][it->second.second].geno_pl[1] = PL1;
	    single_marker[it->second.first][it->second.second].geno_pl[2] = PL2;
	  }

	  //Record the DP info if DP index in FORMAT is >0
	  if(DP_idx>0) {
	    if (std::string((*indv).get(DP_idx, &missing).toStr()).compare(".")==0) single_marker[it->second.first][it->second.second].missing_dp = true;
	    else {
		 single_marker[it->second.first][it->second.second].missing_dp = false;
		 single_marker[it->second.first][it->second.second].dp = (*indv).get(DP_idx, &missing).toInt();
		}
          }
        } // end of reading VCF info for all samples

      //Find the left and right markers with IBD distribution
      if(lpos<0 && rpos<0) { rpos = pm->var[0].pos; ridx=0; }
      if(single_var.pos>=rpos && rpos>0)
        {
	  lpos = rpos;
	  rpos = counter<pm->count ? pm->var[counter+1].pos : -1;
	  lidx = counter;
	  ridx = counter+1;
	  counter++;
        }

      //After loading one variant, do calculation of likelihoods and genotype inference
      //Calculate QUAL from lk based on IBD sharing
       double mono_llk = 0.;
       double poly_llk = 0.;
       double QUAL = 0.;
       single_var.ac=0; single_var.dp=0;

      #pragma omp parallele for firstprivate(lk_l, lk_r, max_lk_value, max_idx),reduction(+:mono_llk, poly_llk, single_var.ac, single_var.dp)
      for(int i=0; i<pm->nFam; i++)
        {
	  lk_l = 0.0; lk_r=-1.0; max_lk_value = 0.0;

          famlk[i].SetGenoPrior(single_var.freq); //Set prior once and will be used in the following calcualtion.

	  if(single_var.pos==lpos)
            {
             max_idx = lidx;
             max_lk_value = famlk[i].CalculateSinglePointLikelihood_IBD(max_idx);
            }
	  else if(single_var.pos>=lpos) // variant is between two variants with IBD probs
	    {
	      if(lpos>0) lk_l = famlk[i].CalculateSinglePointLikelihood_IBD(single_marker[i], famlk[i].marginals[lidx]);
	      if(rpos>0) lk_r = famlk[i].CalculateSinglePointLikelihood_IBD(single_marker[i], famlk[i].marginals[ridx]);

	      if(lk_l > lk_r) { max_idx = lidx; max_lk_value=lk_l; }
	      else { max_idx = ridx; max_lk_value = lk_r; }
	    }
	  else error("Probelm: this should not appear!\n");

	  max_lk_idx[i]=max_idx; //record the loc idx for max lk for variants between two map locations
	  max_lk[i] = max_lk_value;

       if(!phase_marginal && !phase_best)
           famlk[i].CalculateGenoPosteriors(single_marker[i], famlk[i].marginals[max_idx]);
       else
         if(phase_best)
           famlk[i].CalculatePhasedGeno_best(single_marker[i], famlk[i].best_path[max_idx]);
         else
           famlk[i].CalculatePhasedGeno_marginal(single_marker[i], famlk[i].marginals[max_idx]);

	// Update likelihood, AC and DP of the variant across multiple familes
	mono_llk += famlk[i].CalculateRefLogLikelihood(single_marker[i]);
	poly_llk += log10(max_lk_value);
        single_var.ac += famlk[i].ac;
	single_var.dp += famlk[i].dp;

        } //end of for loop for processing genotype inference of all families

       // Calculate QUAL
       std::vector<double> var_llk;
       var_llk.push_back(mono_llk + log10(1-prior) );
       var_llk.push_back(poly_llk + log10(snv_prior) ); //Here the prior is only for the specific REF and ALT pair
       double posterior = MathUtils::CalculatePosteriorFromLLK(var_llk, 1);
       QUAL = -10*log10(1-posterior);
       if(QUAL>100) QUAL=100.;
       single_var.qual = QUAL;

       // Construct INFO and FORMAT fields
       String info = "AF="; info += (1-single_var.freq); info+=";AC="; info+=single_var.ac; info+=";DP="; info+=single_var.dp;
       single_var.info = info.c_str();

        if(phase_best) single_var.format = "GT:GQ:DP:BI:PL";
        else single_var.format = "GT:GQ:DP:MI:PL";

       if(!phase_marginal && !phase_best)
       {
       //Here we cannot do parallelization due to the outptu order but anyway it should be quick
        for(int i=0; i<pm->nFam; i++)
	  OutputVCF(famlk[i], single_var, fh, i, max_lk_idx[i]);
       }
       else
       {
        for(int i=0; i<pm->nFam; i++)
	  OutputVCF_phased(famlk[i], single_var, fh, i, max_lk_idx[i]);
       }

       fprintf(fh, "\n"); //Done with processing one variant for all families and go to the next one

    if(cnt++>0 && cnt%10000==0) printf("Finishing %d varints to output to VCF\n", cnt);
    } //end of while loop for all variants

  fclose(fh);
}

void PedSeq::OutputVCF(FamilyLikelihoodLG &fam, VarInfo &vr, FILE *fh, int famidx, int max_loc_idx)
{
  if(famidx==0) fprintf(fh, "%s%s\t%d\t%s\t%s\t%s\t%.6f\t%s\t%s\t%s", chr_prefix.c_str(), vr.chr.c_str(), vr.pos, vr.id.c_str(), vr.ref.c_str(), vr.alt.c_str(), (vr.qual), vr.filter.c_str(), vr.info.c_str(), vr.format.c_str());

  for(int i=0; i<fam.im->famSize; i++)
  {
     fam_pid_index_pair.first = famidx;
     fam_pid_index_pair.second = i;

     if(output_all==false && pm->fam_pid_index_pair_with_data.count(fam_pid_index_pair)==0) continue;

     if(single_marker[famidx][i].missing_pl) sprintf(pl_str, ".,.,.");
     else sprintf(pl_str, "%d,%d,%d", single_marker[famidx][i].geno_pl[0], single_marker[famidx][i].geno_pl[1],single_marker[famidx][i].geno_pl[2]);

     if(DP_idx>0) {
       if(single_marker[famidx][i].missing_dp) sprintf(dp_str, ".");
       else sprintf(dp_str, "%d", single_marker[famidx][i].dp);

       //fprintf(fh, "\t%s:%d:%s:%d|%d:%d|%d:%s",geno_label[fam.best_geno_idx[i]].c_str(), fam.geno_qual[i], dp_str, fam.im->inherit_matrix[fam.best_marginal[max_loc_idx]][2*i]+1, fam.im->inherit_matrix[fam.best_marginal[max_loc_idx]][2*i+1]+1, fam.im->inherit_matrix[fam.best_path[max_loc_idx]][2*i]+1, fam.im->inherit_matrix[fam.best_path[max_loc_idx]][2*i+1]+1, pl_str);
       fprintf(fh, "\t%s:%d:%s:%d|%d:%s",geno_label[fam.best_geno_idx[i]].c_str(), fam.geno_qual[i], dp_str, fam.im->inherit_matrix[fam.best_marginal[max_loc_idx]][2*i]+1, fam.im->inherit_matrix[fam.best_marginal[max_loc_idx]][2*i+1]+1, pl_str);
     }
     else fprintf(fh, "\t%s:%d:%s",geno_label[fam.best_geno_idx[i]].c_str(), fam.geno_qual[i], pl_str);
  }
  //fflush(fh);
}

void PedSeq::OutputVCF_phased(FamilyLikelihoodLG &fam, VarInfo &vr, FILE *fh, int famidx, int max_loc_idx)
{
  if(famidx==0) fprintf(fh, "%s%s\t%d\t%s\t%s\t%s\t%.6f\t%s\t%s\t%s", chr_prefix.c_str(), vr.chr.c_str(), vr.pos, vr.id.c_str(), vr.ref.c_str(), vr.alt.c_str(), (vr.qual), vr.filter.c_str(), vr.info.c_str(), vr.format.c_str());

  for(int i=0; i<fam.im->famSize; i++)
  {
     fam_pid_index_pair.first = famidx;
     fam_pid_index_pair.second = i;

     if(output_all==false && pm->fam_pid_index_pair_with_data.count(fam_pid_index_pair)==0) continue;

     if(single_marker[famidx][i].missing_pl) sprintf(pl_str, ".,.,.");
     else sprintf(pl_str, "%d,%d,%d", single_marker[famidx][i].geno_pl[0], single_marker[famidx][i].geno_pl[1],single_marker[famidx][i].geno_pl[2]);

     if(DP_idx>0) {
       if(single_marker[famidx][i].missing_dp) sprintf(dp_str, ".");
       else sprintf(dp_str, "%d", single_marker[famidx][i].dp);

     if(phase_best)
       fprintf(fh, "\t%s:%d:%s:%d|%d:%s", fam.best_phased_geno_label[i].c_str(), int(fam.phased_geno_qual), dp_str, fam.im->inherit_matrix[fam.best_path[max_loc_idx]][2*i]+1, fam.im->inherit_matrix[fam.best_path[max_loc_idx]][2*i+1]+1, pl_str);
     else
       fprintf(fh, "\t%s:%d:%s:%d|%d:%s", fam.best_phased_geno_label[i].c_str(), int(fam.phased_geno_qual), dp_str, fam.im->inherit_matrix[fam.best_marginal[max_loc_idx]][2*i]+1, fam.im->inherit_matrix[fam.best_marginal[max_loc_idx]][2*i+1]+1, pl_str);
     }
     else fprintf(fh, "\t%s:%d:%s", fam.best_phased_geno_label[i].c_str(), fam.geno_qual[i], pl_str);
  }
  //fflush(fh);
}
